%% Pre-processing data
clc; clear
K = 2; L = 3; N_Cu = 2; N_Cm = 2; missing_rate = 0.8;
Nu = 300; Nm = 500;

Cu_gen_all = binornd(1,1/2,N_Cu,Nu);
Cm_gen_all = binornd(1,1/2,N_Cm,Nm);
Cu_gen = zeros(N_Cu,K,Nu);
for ii = 1:N_Cu
    for jj = 1:K
        for kk = 1:Nu
            Cu_gen(ii,jj,kk) = Cu_gen_all(ii,kk);
        end
    end
end
Cm_gen = zeros(N_Cm,K,Nu);
for ii = 1:N_Cm
    for jj = 1:K
        for kk = 1:Nm
            Cm_gen(ii,jj,kk) = Cm_gen_all(ii,kk);
        end
    end
end
La_gen = binornd(0,1,N_Cu,K,Nu);
Lb_gen = binornd(0,1,N_Cm,K,Nm);
Lc_gen = rand(L+N_Cu+N_Cm,K);
Ld_gen = rand(L+N_Cu+N_Cm,K);
Lu_gen = normrnd(0,1,L,K,Nu)/10; % users latent factor matrix
Lm_gen = normrnd(0,1,L,K,Nm)/10; % movies latent factor matrix

N = 0;
for ii = 1:Nu*Nm
    if rand() > missing_rate
        N = N + 1;
        all_users(N) = floor(ii/Nm)+1;
        all_movies(N) = rem(ii,Nm)+1;
        dist = exp(sum(([Lu_gen(:,:,all_users(N)); Cu_gen(:,:,all_users(N)); La_gen(:,:,all_users(N))] + Lc_gen) .* ([Lm_gen(:,:,all_movies(N)); Lb_gen(:,:,all_movies(N)); Cm_gen(:,:,all_movies(N))] + Ld_gen),1));
        dist = dist/sum(dist);
        all_ratings(N) = sum((1:K) .* mnrnd(1,dist));
    end
end
order = randperm(N);
all_users = all_users(order)';
all_movies = all_movies(order)';
all_ratings = all_ratings(order)';



%%%%%%%%%%%% training data %%%%%%%%%%%%
train_ratio = 0.6;
tune_ratio = 0.15;
N_train = floor(N * train_ratio);
N_tune = floor(N * tune_ratio);

lrn_users = all_users(1:N_train,:); % training data
lrn_movies = all_movies(1:N_train,:);
lrn_ratings = all_ratings(1:N_train,:);

tune_users = all_users(N_train+1:N_train+N_tune,:); % tuning data
tune_movies = all_movies(N_train+1:N_train+N_tune,:);
tune_ratings = all_ratings(N_train+1:N_train+N_tune,:);

test_users = all_users(N_train+N_tune+1:N,:); % testing data
test_movies = all_movies(N_train+N_tune+1:N,:);
test_ratings = all_ratings(N_train+N_tune+1:N,:);



%%%%%%%%%%%% construct maps %%%%%%%%%%%%
u2ml = cell(Nu,1); % map from a user to rating movies
m2ul = cell(Nm,1); % map from a movie to rated users
for ii = 1:Nu
    u2ml{ii} = find(lrn_users == ii);
end
for jj = 1:Nm
    m2ul{jj} = find(lrn_movies == jj);
end



%%%%%%%%%% save data %%%%%%%%%%%%%
dlmwrite('train',[lrn_users, lrn_movies, lrn_ratings]);
dlmwrite('tune',[tune_users, tune_movies, tune_ratings]);
dlmwrite('test',[test_users, test_movies, test_ratings]);
save('sim.mat')